from types import MethodType, FunctionType
from typing import Union, Optional, Callable

from redis import Redis

from .cache_factory import CacheFactory
from ..caches.cache import Cache
from ..caches.redis_cache import RedisCache


class RedisCacheFactory(CacheFactory):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        password: Optional[str] = None,
        decorator: Optional[Callable[[Cache], Cache]] = None,
    ) -> None:
        super().__init__(decorator)
        self.host = host
        self.port = port
        self.password = password
        self._index = Redis(host, port, db=0, password=password)

    def _create(self, func: Union[MethodType, FunctionType]) -> RedisCache:
        key = self._func_to_key(func)
        db = self._index.get(key)
        if db is None:
            used = {int(self._index.get(k)) for k in self._index.keys("*")}
            db = min({i + 1 for i in range(len(used) + 1)}.difference(used))
            self._index.set(key, db)
        return RedisCache(
            host=self.host, port=self.port, db=db, password=self.password,
        )



from abc import ABC, abstractmethod
from types import MethodType, FunctionType
from typing import Union, Optional, Callable

from ..caches.cache import Cache


class CacheFactory(ABC):
    def __init__(self, decorator: Optional[Callable[[Cache], Cache]]) -> None:
        super().__init__()
        self._decorator = decorator

    def create(self, func: Union[MethodType, FunctionType]) -> Cache:
        result = self._create(func)
        if self._decorator is not None:
            result = self._decorator(result)
        return result

    @abstractmethod
    def _create(self, func: Union[MethodType, FunctionType]) -> Cache:
        pass

    @staticmethod
    def _func_to_key(func: Union[MethodType, FunctionType]) -> str:
        return f"{func.__module__}.{func.__qualname__}"



from time import time
from typing import AnyStr

from .cache import Cache
from ..constants import NOT_FOUND
from ..errors import CachingError
from ..postgresql_key_value_adapter import (
    PostgresqlKeyValueAdapter,
    ConnectionPool,
)


class PostgresqlCache(Cache):
    def __init__(self, connection_pool: ConnectionPool, table: str) -> None:
        super().__init__()
        self._adapter = PostgresqlKeyValueAdapter(connection_pool, table)

    def get(self, key: str):
        t = self._get_curr_time()
        try:
            value = self._adapter.get(key, t)
            if value is None:
                return NOT_FOUND
            return value
        except Exception as e:
            raise CachingError(f"Could not get {key}", exc=e)

    def set(self, key: str, value: AnyStr, expiration: int) -> None:
        if expiration:
            expiration = self._get_curr_time() + expiration
        try:
            self._adapter.set(key, value, expiration)
        except Exception as e:
            raise CachingError(f"Could not set {key} as {value}", exc=e)

    @staticmethod
    def _get_curr_time():
        return round(time() * 1000)



from functools import partial, wraps
from inspect import getsource, iscoroutinefunction, getfullargspec
from types import MethodType
from typing import Optional, Callable, Any, Dict, List, cast

from thornfield.caches.cache import Cache
from .caching_data import CachingData
from .constants import NOT_FOUND
from .errors import CachingError
from .typing import NotCached, Cached, NormalCallable

_CACHE_ATTR = "cache"
_CACHING_DATA_ATTR = "caching_data"


class Cacher:
    def __init__(self, cache_impl: Optional[Callable[[NormalCallable], Cache]]) -> None:
        """

        :param cache_impl: An optional factory function that gets
            the cached method and returns an implementation of ``Cache``.
        """
        super().__init__()
        self._cache_impl = cache_impl

    def cached(
        self,
        cache: Optional[Cache] = None,
        validator: Optional[Callable[[Any], bool]] = None,
        expiration: int = 0,
    ):
        """
        :param cache: The ``Cache`` to use. If ``None``, ``self.cache_impl`` is called to create one.
        :param validator: A ``callable`` that will be called on the return value of the cached function.
            The value will be cached only if ``validator`` returns ``True``.
        :param expiration: Expiration time for each key, in milliseconds.
        """
        if isinstance(cache, Callable):
            return self._cached(cast(NormalCallable, cache), None, None, 0)
        return partial(
            self._cached, cache=cache, validator=validator, expiration=expiration
        )

    def cache_method(
        self,
        method: MethodType,
        cache: Optional[Cache] = None,
        validator: Optional[Callable[[Any], bool]] = None,
        expiration: int = 0,
        use_base_method: bool = True,
    ):
        """

        :param method: The method to cache.
        :param cache: The ``Cache`` to use. If ``None``, ``self.cache_impl`` is called to create one.
        :param validator: A ``callable`` that will be called on the return value of the cached function.
            The value will be cached only if ``validator`` returns ``True``.
        :param expiration: Expiration time for each key, in milliseconds.
        :param use_base_method: If ``method`` is an implementation of a base method,
            whether to pass to ``self.cache_impl`` it or the base method.
        :return:
        """
        func_passed_to_cache = None
        if use_base_method:
            func_passed_to_cache = next(
                getattr(c, method.__name__)
                for c in reversed(method.__self__.__class__.mro())
                if hasattr(c, method.__name__)
            )
        new_method = self._cached(
            method.__func__,
            cache=cache,
            validator=validator,
            expiration=expiration,
            func_passed_to_cache=func_passed_to_cache,
        )
        setattr(
            method.__self__, method.__name__, MethodType(new_method, method.__self__)
        )

    def get_cached_result(self, func: NormalCallable, *args, **kwargs) -> Any:
        wrapped = getattr(func, "__wrapped__", None)
        caching_data: Optional[CachingData] = getattr(func, _CACHING_DATA_ATTR, None)
        if wrapped is None or caching_data is None:
            return NOT_FOUND

        is_instance_func = args and hasattr(args[0], func.__name__)
        func_args = caching_data.func_args
        if len(args) < len(func_args) and func_args[0] == "self":
            func_args = func_args[1:]
        key = self._get_key(
            is_instance_func,
            func_args,
            args,
            kwargs,
            caching_data.func_defaults,
            caching_data.func_annotations,
        )
        func_cache: Optional[Cache] = getattr(wrapped, _CACHE_ATTR, None)
        if func_cache is None:
            func_cache = caching_data.cache or self._cache_impl(
                caching_data.func_passed_to_cache or func
            )
            setattr(wrapped, _CACHE_ATTR, func_cache)
        return func_cache.get(key)

    def _cached(
        self,
        func: NormalCallable,
        cache: Optional[Cache],
        validator: Optional[Callable[[Any], bool]],
        expiration: int,
        func_passed_to_cache: Optional[NormalCallable] = None,
    ):
        if cache is None and self._cache_impl is None:
            raise CachingError("No cache and no cache creator provided.")

        spec = getfullargspec(func)
        func_args = spec.args
        func_annotations = spec.annotations
        if spec.defaults:
            func_defaults = dict(zip(func_args[-len(spec.defaults) :], spec.defaults))
        else:
            func_defaults = {}

        def _x(*args, **kwargs):
            is_instance_func = args and hasattr(args[0], func.__name__)
            key = self._get_key(
                is_instance_func,
                func_args,
                args,
                kwargs,
                func_defaults,
                func_annotations,
            )
            func_cache = getattr(func, _CACHE_ATTR, None)
            if func_cache is None:
                func_cache = cache or self._cache_impl(func_passed_to_cache or func)
                setattr(func, _CACHE_ATTR, func_cache)
            result = func_cache.get(key)
            if result is NOT_FOUND:
                result = func(*args, **kwargs)
                if validator is None or validator(result):
                    func_cache.set(key, result, expiration)
            return result

        source = self._get_inner_code(getsource(_x), func)
        g = dict(globals())
        g.update(locals())
        exec(compile(source, "", "exec"), g, locals())
        inner = locals()["x"]
        inner.caching_data = CachingData(
            func_args=func_args,
            func_annotations=func_annotations,
            func_defaults=func_defaults,
            cache=cache,
            func_passed_to_cache=func_passed_to_cache,
        )
        return wraps(func)(inner)

    @classmethod
    def _get_inner_code(cls, base, func):
        source = base[base.index("(") :]
        if iscoroutinefunction(func):
            source = f"async def x {source}"
            i = source.index("func(")
            source = f"{source[:i]}await {source[i:]}"
        else:
            source = f"def x {source}"
        return source

    @classmethod
    def _get_key(
        cls,
        is_instance_func: bool,
        func_args: List[str],
        args: tuple,
        kwargs: dict,
        defaults: Dict[str, Any],
        annotations: Dict[str, Any],
    ) -> tuple:
        key_args = func_args[1:] if is_instance_func else func_args
        if any(a is NotCached for a in annotations.values()):
            key_args = filter(lambda a: annotations.get(a) is not NotCached, key_args)
        elif any(a is Cached for a in annotations.values()):
            key_args = filter(lambda a: annotations.get(a) is Cached, key_args)
        arg_values = dict(zip(func_args, args))
        arg_values.update(kwargs)
        return tuple(
            arg_values[a] if a in arg_values else defaults[a] for a in key_args
        )



from abc import ABC, abstractmethod
from time import time

from .volatile_value import VolatileValue
from ..constants import NOT_FOUND


class Cache(ABC):
    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def set(self, key, value, expiration: int) -> None:
        pass

    @staticmethod
    def _to_volatile(value, expiration: int) -> VolatileValue:
        return VolatileValue(value, round(time() * 1000) + expiration)

    @staticmethod
    def _from_volatile(value: VolatileValue):
        if time() * 1000 <= value.expiration:
            return value.value
        return NOT_FOUND



import re
from types import MethodType, FunctionType
from typing import Union, Optional, Callable

from .cache_factory import CacheFactory
from ..caches.cache import Cache
from ..caches.postgresql_cache import PostgresqlCache
from ..postgresql_key_value_adapter import (
    PostgresqlKeyValueAdapter,
    ConnectionPool,
)


class PostgresqlCacheFactory(CacheFactory):
    def __init__(
        self,
        connection_pool: ConnectionPool,
        index_table: str = "_index",
        decorator: Optional[Callable[[Cache], Cache]] = None,
    ) -> None:
        super().__init__(decorator)
        self.connection_pool = connection_pool
        self.index_table = index_table
        self._pkv_adapter = None

    @property
    def _adapter(self) -> PostgresqlKeyValueAdapter:
        if self._pkv_adapter is None:
            self._pkv_adapter = PostgresqlKeyValueAdapter(
                self.connection_pool, self.index_table, ts_col=None
            )
        return self._pkv_adapter

    def _create(self, func: Union[MethodType, FunctionType]) -> PostgresqlCache:
        table = self._normalize_table_name(self._func_to_key(func))
        return PostgresqlCache(connection_pool=self.connection_pool, table=table)

    def _normalize_table_name(self, table_name: str) -> str:
        table_name = table_name.lower()
        if len(table_name) > 63:
            n = self._adapter.get(table_name)
            if not n:
                used = {int(self._adapter.get(k)) for k in self._adapter.keys()}
                n = min({i + 1 for i in range(len(used) + 1)}.difference(used))
                self._adapter.set(table_name, str(n))
            return f"cache_{n}"
        else:
            return re.sub(r"[^\w]", "_", table_name)



from typing import Optional, AnyStr

from .cache import Cache
from ..constants import NOT_FOUND
from ..errors import CachingError

try:
    from redis import Redis
except ModuleNotFoundError:
    Redis = None


class RedisCache(Cache):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__()
        if Redis is None:
            raise CachingError('Package "redis" is not installed')
        self._redis = Redis(
            host=host,
            port=port,
            db=db,
            password=password,
            decode_responses=True,
            **kwargs,
        )

    def get(self, key: str) -> AnyStr:
        try:
            value = self._redis.get(key)
            if value is None:
                return NOT_FOUND
            return value
        except Exception as e:
            raise CachingError(f"Could not get {key}", exc=e)

    def set(self, key: str, value: AnyStr, expiration: int) -> None:
        try:
            self._redis.set(key, value, px=expiration or None)
        except Exception as e:
            raise CachingError(f"Could not set {key} as {value}", exc=e)



from types import FunctionType, MethodType
from typing import TypeVar, Type, Union

NormalCallable = Union[FunctionType, MethodType]
T = TypeVar("T")


class _Cached:
    def __getitem__(self, item: Type[T]) -> Type[T]:
        return self


class _NotCached:
    def __getitem__(self, item: Type[T]) -> Type[T]:
        return self


Cached = _Cached()
NotCached = _NotCached()



from typing import Callable, AnyStr, Optional
from zlib import compress as default_compress, decompress as default_decompress

from .cache import Cache
from ..constants import NOT_FOUND


class CacheCompressionDecorator(Cache):
    def __init__(
        self,
        cache: Cache,
        compress: Optional[Callable[[str], AnyStr]] = ...,
        decompress: Optional[Callable[[AnyStr], str]] = ...,
    ) -> None:
        super().__init__()
        self._cache = cache
        if compress is None:
            self._compress = self._noop
        elif compress is ...:
            self._compress = self._default_compress
        else:
            self._compress = compress

        if decompress is None:
            self._decompress = self._noop
        elif decompress is ...:
            self._decompress = self._default_decompress
        else:
            self._decompress = decompress

    def get(self, key):
        value = self._cache.get(key)
        return value if value is NOT_FOUND else self._decompress(value)

    def set(self, key, value, expiration: int) -> None:
        self._cache.set(key, self._compress(value), expiration)

    @staticmethod
    def _noop(x):
        return x

    @staticmethod
    def _default_compress(obj: str) -> bytes:
        return default_compress(obj.encode("UTF-8"))

    @staticmethod
    def _default_decompress(data: bytes) -> str:
        return default_decompress(data).decode("UTF-8")



from typing import Any

from .cache import Cache
from .volatile_value import VolatileValue
from ..constants import NOT_FOUND


class MemoryCache(Cache):
    def __init__(self) -> None:
        super().__init__()
        self._cache = {}

    def get(self, key: str) -> Any:
        if key not in self._cache:
            return NOT_FOUND
        value = self._cache[key]
        if isinstance(value, VolatileValue):
            return self._from_volatile(value)
        return value

    def set(self, key: str, value: Any, expiration: int) -> None:
        if expiration:
            value = self._to_volatile(value, expiration)
        self._cache[key] = value



import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="thornfield",
    version="1.5.0",
    author="Dror A. Vinkler",
    description="Advanced caching in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/drorvinkler/thornfield",
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    license='MIT',
)



from enum import Enum, auto
from typing import Callable, Optional, List, Union, AnyStr

try:
    from psycopg2.pool import AbstractConnectionPool
except ImportError:
    AbstractConnectionPool = None

ConnectionPool = Union[AbstractConnectionPool, Callable[[], AbstractConnectionPool]]


class FetchAmount(Enum):
    ONE = auto()
    ALL = auto()
    ZERO = auto()


class ConnectionWrapper:
    def __init__(
        self, connection, release_callback: Callable[["ConnectionWrapper"], None]
    ) -> None:
        super().__init__()
        self._connection = connection
        self._callback = release_callback

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._callback(self._connection)

    def execute_query(self, query: str, fetch: FetchAmount, *params: str):
        with self._connection.cursor() as cursor:
            cursor.execute(query, params)
            if fetch is FetchAmount.ONE:
                return cursor.fetchone() if cursor.rowcount > 0 else None
            elif fetch is FetchAmount.ALL:
                return cursor.fetchall()
        self._connection.commit()


class ConnectionPoolWrapper:
    def __init__(self, pool: ConnectionPool) -> None:
        super().__init__()
        self._raw_pool = pool

    @property
    def _pool(self):
        if isinstance(self._raw_pool, Callable):
            self._raw_pool = self._raw_pool()
        return self._raw_pool

    def getconn(self) -> ConnectionWrapper:
        return ConnectionWrapper(self._pool.getconn(), self.putconn)

    def putconn(self, conn):
        self._pool.putconn(conn)


class PostgresqlKeyValueAdapter:
    def __init__(
        self,
        connection_pool: ConnectionPool,
        table_name: str,
        key_col: str = "key",
        value_col: str = "value",
        ts_col: Optional[str] = "ts",
    ) -> None:
        super().__init__()
        self._pool = ConnectionPoolWrapper(connection_pool)
        self._table = table_name
        self._key_col = key_col
        self._value_col = value_col
        self._ts_col = ts_col
        with self._pool.getconn() as connection:
            self._table_exists = self._exists(
                connection, "information_schema.tables", "table_name", self._table
            )

    def set(self, key: str, value: AnyStr, ts: Optional[int] = None):
        if self._ts_col:
            assert ts is not None
        if not self._table_exists:
            self._create_table_if_not_exists(isinstance(value, bytes))
            self._table_exists = True
        with self._pool.getconn() as connection:
            if self._exists(connection, self._table, self._key_col, key):
                self._update(connection, key, value, ts)
            else:
                self._add(connection, key, value, ts)

    def get(self, key: str, min_ts: Optional[int] = None) -> Optional[AnyStr]:
        if not self._table_exists:
            return None

        query = f"select {self._value_col} from {self._table} where {self._key_col}=%s"
        if min_ts is not None:
            assert self._ts_col
            query += f" and ({self._ts_col}>{min_ts} or {self._ts_col}=0)"
        with self._pool.getconn() as connection:
            result = connection.execute_query(query, FetchAmount.ONE, key)
        return result[0] if result else None

    def keys(self) -> List[str]:
        if not self._table_exists:
            return []

        with self._pool.getconn() as connection:
            result = connection.execute_query(
                f"select {self._key_col} from {self._table}", FetchAmount.ALL
            )
        return [t[0] for t in result]

    def _create_table_if_not_exists(self, binary: bool):
        with self._pool.getconn() as connection:
            exists = self._exists(
                connection, "information_schema.tables", "table_name", self._table
            )
            if exists:
                return
            value_col_type = "bytes" if binary else "text"
            structure = f"id serial primary key, {self._key_col} text unique, {self._value_col} {value_col_type}"
            if self._ts_col:
                structure += f", {self._ts_col} bigint"
            connection.execute_query(
                f"create table {self._table} ({structure})", FetchAmount.ZERO
            )

    def _update(self, connection: ConnectionWrapper, key, value, ts):
        values_str = f"{self._value_col}=%s"
        if self._ts_col:
            values_str += f", {self._ts_col}={ts}"
        connection.execute_query(
            f"update {self._table} set {values_str} where {self._key_col}=%s",
            FetchAmount.ZERO,
            value,
            key,
        )

    def _add(self, connection: ConnectionWrapper, key, value, ts):
        columns = f"{self._key_col}, {self._value_col}"
        values = "%s, %s"
        if self._ts_col:
            columns += f", {self._ts_col}"
            values += f", {ts}"
        connection.execute_query(
            f"insert into {self._table} ({columns}) values ({values})",
            FetchAmount.ZERO,
            key,
            value,
        )

    @classmethod
    def _exists(
        cls, connection: ConnectionWrapper, table: str, column: str, value: str
    ) -> bool:
        exists = connection.execute_query(
            f"select exists(select * from {table} where {column}=%s)",
            FetchAmount.ONE,
            value,
        )
        return exists[0]



import json
from typing import Optional, Callable, Any

from .cache import Cache
from ..constants import NOT_FOUND
from ..errors import CachingError

try:
    from yasoo import serialize, deserialize
except ImportError:
    serialize = deserialize = None


class CacheSerializationDecorator(Cache):
    def __init__(
        self,
        cache: Cache,
        serializer: Optional[Callable[[Any], str]] = ...,
        deserializer: Optional[Callable[[Optional[str]], Any]] = ...,
    ) -> None:
        super().__init__()
        self._cache = cache

        if (serializer is ... or deserializer is ...) and serialize is None:
            raise CachingError(
                'Package "yasoo" is not installed and no de/serializer passed'
            )
        if serializer is None:
            self._serialize = self._noop
        elif serializer is ...:
            self._serialize = self._default_serialize
        else:
            self._serialize = serializer

        if deserializer is None:
            self._deserialize = self._noop
        elif deserializer is ...:
            self._deserialize = self._default_deserialize
        else:
            self._deserialize = deserializer

    def get(self, key):
        value = self._cache.get(self._serialize(key))
        return value if value is NOT_FOUND else self._deserialize(value)

    def set(self, key, value, expiration: int) -> None:
        self._cache.set(self._serialize(key), self._serialize(value), expiration)

    @staticmethod
    def _noop(x):
        return x

    @staticmethod
    def _default_serialize(obj) -> str:
        return json.dumps(serialize(obj, preserve_iterable_types=True))

    @staticmethod
    def _default_deserialize(data: Optional[str]):
        if data is None:
            return data
        return deserialize(json.loads(data))



from typing import Optional


class CachingError(BaseException):
    def __init__(self, message, exc: Optional[BaseException] = None) -> None:
        super().__init__(message)
        self.exc = exc



from dataclasses import dataclass
from typing import List, Dict, Any, Optional

from thornfield.caches.cache import Cache
from thornfield.typing import NormalCallable


@dataclass
class CachingData:
    func_args: List[str]
    func_annotations: Dict[str, Any]
    func_defaults: Dict[str, Any]
    cache: Optional[Cache]
    func_passed_to_cache: Optional[NormalCallable]
